@@ -1,6 +1,6 @@
 /*
   This file is part of TALER
-  (C) 2014, 2015, 2016 INRIA
+  (C) 2014-2017 INRIA
 
   TALER is free software; you can redistribute it and/or modify it under the
   terms of the GNU Affero General Public License as published by the Free Software
@@ -11,7 +11,7 @@
   A PARTICULAR PURPOSE.  See the GNU General Public License for more details.
 
   You should have received a copy of the GNU General Public License along with
-  TALER; see the file COPYING.  If not, If not, see <http://www.gnu.org/licenses/>
+  TALER; see the file COPYING.  If not, see <http://www.gnu.org/licenses/>
 */
 /**
  * @file merchant/backend/taler-merchant-httpd.c
@@ -35,47 +35,37 @@
 #include "taler-merchant-httpd_mhd.h"
 #include "taler-merchant-httpd_auditors.h"
 #include "taler-merchant-httpd_exchanges.h"
-#include "taler-merchant-httpd_contract.h"
+#include "taler-merchant-httpd_proposal.h"
 #include "taler-merchant-httpd_pay.h"
-#include "taler-merchant-httpd_track.h"
-
+#include "taler-merchant-httpd_track-transaction.h"
+#include "taler-merchant-httpd_track-transfer.h"
+#include "taler-merchant-httpd_history.h"
+#include "taler-merchant-httpd_map.h"
 
 /**
  * Backlog for listen operation on unix-domain sockets.
  */
 #define UNIX_BACKLOG 500
 
+/**
+ * Hashmap pointing at merchant instances by 'id'. An 'id' is
+ * just a string that identifies a merchant instance. When a frontend
+ * needs to specify an instance to the backend, it does so by 'id'
+ */
+struct GNUNET_CONTAINER_MultiHashMap *by_id_map;
 
 /**
- * Our wire format details in JSON format (with salt).
+ * Hashmap pointing at merchant instances by public key. This map
+ * is mainly used to check whether there is more than one instance
+ * using the same key
  */
-struct json_t *j_wire;
-
-/**
- * Hash of our wire format details as given in #j_wire.
- */
-struct GNUNET_HashCode h_wire;
-
-/**
- * Merchant's private key
- */
-struct GNUNET_CRYPTO_EddsaPrivateKey *privkey;
-
-/**
- * Merchant's public key
- */
-struct TALER_MerchantPublicKeyP pubkey;
+struct GNUNET_CONTAINER_MultiHashMap *by_kpub_map;
 
 /**
  * The port we are running on
  */
 static long long unsigned port;
 
-/**
- * File holding the merchant's private key
- */
-static char *keyfile;
-
 /**
  * This value tells the exchange by which date this merchant would like
  * to receive the funds for a deposited payment
@@ -83,9 +73,20 @@ static char *keyfile;
 struct GNUNET_TIME_Relative wire_transfer_delay;
 
 /**
- * Which currency is supported by this merchant?
+ * Default maximum wire fee to assume, unless stated differently in the proposal
+ * already.
  */
-char *TMH_merchant_currency_string;
+struct TALER_Amount default_max_wire_fee;
+
+/**
+ * Default factor for wire fee amortization.
+ */
+unsigned long long default_wire_fee_amortization;
+
+/**
+ * Should a "Connection: close" header be added to each HTTP response?
+ */
+int TMH_merchant_connection_close;
 
 /**
  * Task running the HTTP server.
@@ -174,25 +175,36 @@ url_handler (void *cls,
       { "/", MHD_HTTP_METHOD_GET, "text/plain",
         "Hello, I'm a merchant's Taler backend. This HTTP server is not for humans.\n", 0,
         &TMH_MHD_handler_static_response, MHD_HTTP_OK },
-      { "/contract", MHD_HTTP_METHOD_POST, "application/json",
+      { "/proposal", MHD_HTTP_METHOD_POST, "application/json",
         NULL, 0,
-        &MH_handler_contract, MHD_HTTP_OK },
-      { "/contract", NULL, "text/plain",
-        "Only POST is allowed", 0,
-        &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED },
-
+        &MH_handler_proposal_put, MHD_HTTP_OK },
       { "/pay", MHD_HTTP_METHOD_POST, "application/json",
         NULL, 0,
         &MH_handler_pay, MHD_HTTP_OK },
       { "/pay", NULL, "text/plain",
         "Only POST is allowed", 0,
         &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED },
-      { "/track/deposit", MHD_HTTP_METHOD_GET, "application/json",
+      { "/track/transfer", MHD_HTTP_METHOD_GET, "application/json",
         NULL, 0,
-        &MH_handler_track_deposit, MHD_HTTP_OK},
-      { "/track/deposit", NULL, "text/plain",
+        &MH_handler_track_transfer, MHD_HTTP_OK},
+      { "/track/transfer", NULL, "text/plain",
         "Only GET is allowed", 0,
         &TMH_MHD_handler_static_response, MHD_HTTP_OK},
+      { "/track/transaction", MHD_HTTP_METHOD_GET, "application/json",
+        NULL, 0,
+        &MH_handler_track_transaction, MHD_HTTP_OK},
+      { "/track/transaction", NULL, "text/plain",
+        "Only GET is allowed", 0,
+        &TMH_MHD_handler_static_response, MHD_HTTP_OK},
+      { "/history", MHD_HTTP_METHOD_GET, "text/plain",
+        "Only GET is allowed", 0,
+        &MH_handler_history, MHD_HTTP_OK},
+      { "/proposal", MHD_HTTP_METHOD_GET, "text/plain",
+        "Only GET is allowed", 0,
+        &MH_handler_proposal_lookup, MHD_HTTP_OK},
+      { "/proposal", NULL, "text/plain",
+        "Only GET/POST are allowed", 0,
+        &TMH_MHD_handler_send_json_pack_error, MHD_HTTP_METHOD_NOT_ALLOWED },
 
       {NULL, NULL, NULL, NULL, 0, 0 }
     };
@@ -207,7 +219,7 @@ url_handler (void *cls,
   unsigned int i;
   int ret;
 
-  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+  GNUNET_log (GNUNET_ERROR_TYPE_INFO,
               "Handling request for URL `%s'\n",
               url);
   for (i=0;NULL != handlers[i].url;i++)
@@ -238,6 +250,28 @@ url_handler (void *cls,
 }
 
 
+/**
+ * Callback that frees all the elements in the hashmap
+ *
+ * @param cls closure, NULL
+ * @param key current key
+ * @param value a `struct MerchantInstance`
+ */
+static int
+hashmap_free (void *cls,
+              const struct GNUNET_HashCode *key,
+              void *value)
+{
+  struct MerchantInstance *mi = value;
+
+  json_decref (mi->j_wire);
+  GNUNET_free (mi->id);
+  GNUNET_free (mi->keyfile);
+  GNUNET_free (mi);
+  return GNUNET_YES;
+}
+
+
 /**
  * Shutdown task (magically invoked when the application is being
  * quit)
@@ -252,6 +286,7 @@ do_shutdown (void *cls)
     GNUNET_SCHEDULER_cancel (mhd_task);
     mhd_task = NULL;
   }
+  MH_force_pc_resume ();
   if (NULL != mhd)
   {
     MHD_stop_daemon (mhd);
@@ -264,12 +299,19 @@ do_shutdown (void *cls)
   }
   TMH_EXCHANGES_done ();
   TMH_AUDITORS_done ();
-  if (NULL != keyfile)
-    GNUNET_free (privkey);
-  if (NULL != j_wire)
+
+  GNUNET_CONTAINER_multihashmap_iterate (by_id_map,
+                                         &hashmap_free,
+                                         NULL);
+  if (NULL != by_id_map)
   {
-    json_decref (j_wire);
-    j_wire = NULL;
+    GNUNET_CONTAINER_multihashmap_destroy (by_id_map);
+    by_id_map = NULL;
+  }
+  if (NULL != by_kpub_map)
+  {
+    GNUNET_CONTAINER_multihashmap_destroy (by_kpub_map);
+    by_kpub_map = NULL;
   }
 }
 
@@ -344,72 +386,6 @@ TMH_trigger_daemon ()
 }
 
 
-/**
- * Verify that #j_wire contains a well-formed wire format, and
- * update #h_wire to match it (if successful).
- *
- * @param cfg configuration to use
- * @param allowed which wire format is allowed/expected?
- * @return #GNUNET_OK on success, #GNUNET_SYSERR on error
- */
-static int
-validate_and_hash_wireformat (const struct GNUNET_CONFIGURATION_Handle *cfg,
-                              const char *allowed)
-{
-  struct TALER_WIRE_Plugin *plugin;
-  char *lib_name;
-  int ret;
-
-  (void) GNUNET_asprintf (&lib_name,
-                          "libtaler_plugin_wire_%s",
-                          allowed);
-  plugin = GNUNET_PLUGIN_load (lib_name,
-                               NULL);
-  if (NULL == plugin)
-  {
-    GNUNET_free (lib_name);
-    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                "Wire transfer method `%s' not supported\n",
-                allowed);
-    return GNUNET_NO;
-  }
-  plugin->library_name = lib_name;
-  j_wire = plugin->get_wire_details (plugin->cls,
-                                     cfg,
-                                     "merchant-wireformat");
-  if (NULL == j_wire)
-    ret = GNUNET_SYSERR;
-  else
-    ret = plugin->wire_validate (plugin->cls,
-                                 j_wire,
-                                 NULL);
-  GNUNET_PLUGIN_unload (lib_name,
-                        plugin);
-  GNUNET_free (lib_name);
-  if (GNUNET_YES != ret)
-    return GNUNET_SYSERR;
-  if (GNUNET_SYSERR ==
-      TALER_JSON_hash (j_wire,
-                       &h_wire))
-    return GNUNET_SYSERR;
-  return GNUNET_OK;
-}
-
-
-/**
- * Custom cleanup routine for a `struct PayContext`.
- *
- * @param hc the `struct PayContext` to clean up.
- */
-void
-TMH_json_parse_cleanup (struct TM_HandlerContext *hc)
-{
-  struct TMH_JsonParseContext *jpc = (struct TMH_JsonParseContext *) hc;
-
-  TMH_PARSE_post_cleanup_callback (jpc->json_parse_context);
-}
-
-
 /**
  * Function that queries MHD's select sets and
  * starts the task waiting for them.
@@ -462,6 +438,287 @@ prepare_daemon ()
 }
 
 
+/**
+ * Callback that looks for 'merchant-instance-*' sections,
+ * and populates accordingly each instance's data
+ *
+ * @param cls closure
+ * @section section name this callback gets
+ */
+static void
+instances_iterator_cb (void *cls,
+                       const char *section)
+{
+  char *substr;
+  char *token;
+  char *instance_wiresection;
+  struct MerchantInstance *mi;
+  struct IterateInstancesCls *iic;
+  struct GNUNET_CRYPTO_EddsaPrivateKey *pk;
+  /* used as hashmap keys */
+  struct GNUNET_HashCode h_pk;
+  struct GNUNET_HashCode h_id;
+  json_t *type;
+  char *emsg;
+
+  iic = cls;
+  substr = strstr (section, "merchant-instance-");
+
+  if ( (NULL == substr) ||
+       (NULL != strstr (section,
+                        "merchant-instance-wireformat-")) )
+    return;
+
+  if (substr != section)
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to specify a merchant instance\n");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+
+  /** Get id **/
+  token = strrchr (section, '-');
+  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+              "Extracted token: %s\n",
+              token + 1);
+  mi = GNUNET_new (struct MerchantInstance);
+
+  if (GNUNET_OK !=
+      GNUNET_CONFIGURATION_get_value_filename (iic->config,
+                                               section,
+                                               "KEYFILE",
+                                               &mi->keyfile))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               section,
+                               "KEYFILE");
+    GNUNET_free (mi);
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+
+  if (GNUNET_YES != GNUNET_DISK_file_test (mi->keyfile))
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Merchant private key `%s' does not exist yet, creating it!\n",
+                mi->keyfile);
+  if (NULL ==
+       (pk = GNUNET_CRYPTO_eddsa_key_create_from_file (mi->keyfile)))
+  {
+    GNUNET_break (0);
+    GNUNET_free (mi->keyfile);
+    GNUNET_free (mi);
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+  mi->privkey.eddsa_priv = *pk;
+  GNUNET_CRYPTO_eddsa_key_get_public (pk,
+                                      &mi->pubkey.eddsa_pub);
+  GNUNET_free (pk);
+
+  mi->id = GNUNET_strdup (token + 1);
+  if (0 == strcmp ("default", mi->id))
+    iic->default_instance = GNUNET_YES;
+
+  GNUNET_asprintf (&instance_wiresection,
+                   "merchant-instance-wireformat-%s",
+                   mi->id);
+  mi->j_wire = iic->plugin->get_wire_details (iic->plugin->cls,
+                                              iic->config,
+                                              instance_wiresection);
+  GNUNET_free (instance_wiresection);
+  if ( (NULL == (type = json_object_get (mi->j_wire,
+                                         "type"))) ||
+       (! json_is_string (type)) )
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Malformed wireformat: lacks type\n");
+    iic->ret |= GNUNET_SYSERR;
+  }
+  mi->wire_method = json_string_value (type);
+
+  if (TALER_EC_NONE !=
+      iic->plugin->wire_validate (iic->plugin->cls,
+                                  mi->j_wire,
+                                  NULL,
+                                  &emsg))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Malformed wireformat: %s\n",
+                emsg);
+    GNUNET_free (emsg);
+    iic->ret |= GNUNET_SYSERR;
+  }
+
+  if (GNUNET_YES != TALER_JSON_hash (mi->j_wire,
+                                     &mi->h_wire))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to hash wireformat\n");
+    iic->ret |= GNUNET_SYSERR;
+  }
+  #define EXTRADEBUG
+  #ifdef EXTRADEBUGG
+  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+              "Found wireformat instance:\n");
+              json_dumpf (mi->j_wire, stdout, 0);
+              printf ("\n");
+  #endif
+
+  GNUNET_CRYPTO_hash (mi->id,
+                      strlen (mi->id),
+                      &h_id);
+  GNUNET_CRYPTO_hash (&mi->pubkey.eddsa_pub,
+                      sizeof (struct GNUNET_CRYPTO_EddsaPublicKey),
+                      &h_pk);
+  if (GNUNET_OK !=
+      GNUNET_CONTAINER_multihashmap_put (by_id_map,
+                                         &h_id,
+                                         mi,
+                                         GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_ONLY))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to put an entry into the 'by_id' hashmap\n");
+    iic->ret |= GNUNET_SYSERR;
+  }
+  if (GNUNET_OK !=
+      GNUNET_CONTAINER_multihashmap_put (by_kpub_map,
+                                         &h_pk,
+                                         mi,
+                                         GNUNET_CONTAINER_MULTIHASHMAPOPTION_UNIQUE_ONLY))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to put an entry into the 'by_kpub_map' hashmap\n");
+    iic->ret |= GNUNET_SYSERR;
+  }
+}
+
+
+/**
+ * Lookup a merchant instance by its name.
+ *
+ * @param name name of the instance to resolve
+ * @return NULL if that instance is unknown to us
+ */
+struct MerchantInstance *
+TMH_lookup_instance (const char *name)
+{
+  struct GNUNET_HashCode h_instance;
+
+  GNUNET_CRYPTO_hash (name,
+                      strlen (name),
+                      &h_instance);
+  GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+              "Looking for by-id key %s of '%s' in hashmap\n",
+              GNUNET_h2s (&h_instance),
+              name);
+  /* We're fine if that returns NULL, the calling routine knows how
+     to handle that */
+  return GNUNET_CONTAINER_multihashmap_get (by_id_map,
+                                            &h_instance);
+}
+
+
+/**
+ * Extract merchant instance from the given JSON
+ *
+ * @param json the JSON to inspect; it is not required to
+ * comply with any particular format. It will only be checked
+ * if the field "instance" is there.
+ * @return a pointer to a #struct MerchantInstance. This will be
+ * the 'default' merchant if the frontend did not specif any
+ * "instance" field. The user should not care to free the returned
+ * value, as it is taken from a global array that will be freed
+ * by the general shutdown routine. NULL if the frontend specified
+ * a wrong instance
+ */
+struct MerchantInstance *
+get_instance (struct json_t *json)
+{
+  struct json_t *instance;
+  const char *instance_str;
+
+  if (NULL == (instance = json_object_get (json, "instance")))
+    instance_str = "default";
+  else
+    instance_str = json_string_value (instance);
+  return TMH_lookup_instance (instance_str);
+}
+
+
+/**
+ * Iterate over each merchant instance, in order to populate
+ * each instance's own data
+ *
+ * @param config configuration handle
+ * @param allowed which wire format is allowed/expected?
+ * @return #GNUNET_OK if successful, #GNUNET_SYSERR upon errors
+ * (for example, if no "default" instance is defined)
+ */
+static unsigned int
+iterate_instances (const struct GNUNET_CONFIGURATION_Handle *config,
+                   const char *allowed)
+{
+  struct IterateInstancesCls *iic;
+  char *lib_name;
+
+  (void) GNUNET_asprintf (&lib_name,
+                          "libtaler_plugin_wire_%s",
+                          allowed);
+  iic = GNUNET_new (struct IterateInstancesCls);
+  iic->current_index = 0;
+  iic->config = config;
+  iic->default_instance = GNUNET_NO;
+  iic->plugin = GNUNET_PLUGIN_load (lib_name,
+                                    NULL);
+  if (NULL == iic->plugin)
+  {
+    GNUNET_free (lib_name);
+    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
+                "Wire transfer method `%s' not supported\n",
+                allowed);
+    return GNUNET_SYSERR;
+  }
+  iic->plugin->library_name = lib_name;
+  GNUNET_CONFIGURATION_iterate_sections (config,
+                                         &instances_iterator_cb,
+                                         (void *) iic);
+
+  if (GNUNET_NO == iic->default_instance)
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "No default merchant instance found\n");
+    goto fail;
+  }
+  else
+    GNUNET_log (GNUNET_ERROR_TYPE_DEBUG,
+                "Congratulations, you have a default instance\n");
+
+  if (0 != (GNUNET_SYSERR & iic->ret))
+  {
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+               "At least one instance has not been successfully parsed\n");
+    goto fail;
+  }
+
+  GNUNET_PLUGIN_unload (lib_name,
+                        iic->plugin);
+  GNUNET_free (lib_name);
+  GNUNET_free (iic);
+  return GNUNET_OK;
+
+ fail:
+  do {
+    GNUNET_PLUGIN_unload (lib_name,
+                          iic->plugin);
+    GNUNET_free (lib_name);
+    GNUNET_free (iic);
+    GNUNET_SCHEDULER_shutdown ();
+    return GNUNET_SYSERR;
+  } while (0);
+}
+
+
 /**
  * Main function that will be run by the scheduler.
  *
@@ -478,14 +735,14 @@ run (void *cls,
      const struct GNUNET_CONFIGURATION_Handle *config)
 {
   char *wireformat;
+  int fh;
 
-  wireformat = NULL;
   result = GNUNET_SYSERR;
   GNUNET_SCHEDULER_add_shutdown (&do_shutdown,
                                  NULL);
   GNUNET_assert (GNUNET_OK ==
                  GNUNET_log_setup ("taler-merchant-httpd",
-                                   "INFO",
+                                   "DEBUG",
                                    NULL));
   if (GNUNET_SYSERR ==
       TMH_EXCHANGES_init (config))
@@ -499,15 +756,17 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
-  if (GNUNET_SYSERR ==
-      GNUNET_CONFIGURATION_get_value_string (config,
-                                             "taler",
-                                             "currency",
-                                             &TMH_merchant_currency_string))
+
+  if (NULL ==
+     (by_id_map = GNUNET_CONTAINER_multihashmap_create(1, GNUNET_NO)))
+  {
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+
+  if (NULL ==
+     (by_kpub_map = GNUNET_CONTAINER_multihashmap_create(1, GNUNET_NO)))
   {
-    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                               "taler",
-                               "currency");
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
@@ -524,6 +783,56 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
+
+  if (GNUNET_OK !=
+      TALER_config_get_denom (config,
+                              "merchant",
+                              "DEFAULT_MAX_WIRE_FEE",
+                              &default_max_wire_fee))
+  {
+    char *currency;
+
+    if (GNUNET_OK !=
+        GNUNET_CONFIGURATION_get_value_string (config,
+                                               "taler",
+                                               "CURRENCY",
+                                               &currency))
+    {
+      GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                                 "taler",
+                                 "CURRENCY");
+      GNUNET_SCHEDULER_shutdown ();
+      return;
+    }
+    if (GNUNET_OK !=
+        TALER_amount_get_zero (currency,
+                               &default_max_wire_fee))
+    {
+      GNUNET_log_config_invalid (GNUNET_ERROR_TYPE_ERROR,
+                                 "taler",
+                                 "CURRENCY",
+                                 "Specified value not legal for a Taler currency");
+      GNUNET_SCHEDULER_shutdown ();
+      GNUNET_free (currency);
+      return;
+    }
+    GNUNET_free (currency);
+  }
+
+  if (GNUNET_OK !=
+      GNUNET_CONFIGURATION_get_value_number (config,
+                                             "merchant",
+                                             "DEFAULT_WIRE_FEE_AMORTIZATION",
+                                             &default_wire_fee_amortization))
+  {
+    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                               "merchant",
+                               "DEFAULT_WIRE_FEE_AMORTIZATION");
+    GNUNET_SCHEDULER_shutdown ();
+    return;
+  }
+
+  wireformat = NULL;
   if (GNUNET_OK !=
       GNUNET_CONFIGURATION_get_value_string (config,
                                              "merchant",
@@ -536,41 +845,10 @@ run (void *cls,
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
-  if (GNUNET_OK !=
-      validate_and_hash_wireformat (config,
-                                    wireformat))
-  {
-    GNUNET_SCHEDULER_shutdown ();
-    return;
-  }
+  iterate_instances (config,
+                     wireformat);
   GNUNET_free (wireformat);
 
-  if (GNUNET_OK !=
-      GNUNET_CONFIGURATION_get_value_filename (config,
-                                               "merchant",
-                                               "KEYFILE",
-                                               &keyfile))
-  {
-    GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                               "merchant",
-                               "KEYFILE");
-    GNUNET_SCHEDULER_shutdown ();
-    return;
-  }
-  if (GNUNET_YES != GNUNET_DISK_file_test (keyfile))
-    GNUNET_log (GNUNET_ERROR_TYPE_INFO,
-                "Merchant private key `%s' does not exist yet, creating it!\n",
-                keyfile);
-  if (NULL ==
-      (privkey =
-       GNUNET_CRYPTO_eddsa_key_create_from_file (keyfile)))
-  {
-    GNUNET_break (0);
-    GNUNET_SCHEDULER_shutdown ();
-    return;
-  }
-  GNUNET_CRYPTO_eddsa_key_get_public (privkey,
-                                      &pubkey.eddsa_pub);
   if (NULL ==
       (db = TALER_MERCHANTDB_plugin_load (config)))
   {
@@ -605,41 +883,11 @@ run (void *cls,
       return;
     }
 
-    // FIXME: refactor two calls to MHD_start_daemon
-    // into one, using an options array instead of varags
-
-    if (0 == strcmp (serve_type, "tcp"))
-    {
-
-      if (GNUNET_SYSERR ==
-          GNUNET_CONFIGURATION_get_value_number (config,
-                                                 "merchant",
-                                                 "PORT",
-                                                 &port))
-      {
-        GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
-                                   "merchant",
-                                   "PORT");
-        GNUNET_SCHEDULER_shutdown ();
-        return;
-      }
-
-      mhd = MHD_start_daemon (MHD_USE_SUSPEND_RESUME,
-                              port,
-                              NULL, NULL,
-                              &url_handler, NULL,
-                              MHD_OPTION_NOTIFY_COMPLETED,
-                              &handle_mhd_completion_callback, NULL,
-                              MHD_OPTION_CONNECTION_TIMEOUT,
-                              (unsigned int) 10 /* 10s */,
-                              MHD_OPTION_END);
-    }
-    else if (0 == strcmp (serve_type, "unix"))
+    if (0 == strcmp (serve_type, "unix"))
     {
       struct sockaddr_un *un;
       char *mode;
       struct GNUNET_NETWORK_Handle *nh;
-      int fh;
 
       if (GNUNET_OK !=
           GNUNET_CONFIGURATION_get_value_filename (config,
@@ -703,49 +951,144 @@ run (void *cls,
 
       un = GNUNET_new (struct sockaddr_un);
       un->sun_family = AF_UNIX;
-      strncpy (un->sun_path, serve_unixpath, sizeof (un->sun_path) - 1);
+      strncpy (un->sun_path,
+               serve_unixpath,
+               sizeof (un->sun_path) - 1);
 
       GNUNET_NETWORK_unix_precheck (un);
 
-      if (NULL == (nh = GNUNET_NETWORK_socket_create (AF_UNIX, SOCK_STREAM, 0)))
+      if (NULL == (nh = GNUNET_NETWORK_socket_create (AF_UNIX,
+                                                      SOCK_STREAM,
+                                                      0)))
       {
-        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR, "create(for AF_UNIX)");
+        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                             "socket(AF_UNIX)");
         GNUNET_SCHEDULER_shutdown ();
         return;
       }
-      if (GNUNET_OK != GNUNET_NETWORK_socket_bind (nh, (void *) un, sizeof (struct sockaddr_un)))
+      if (GNUNET_OK !=
+          GNUNET_NETWORK_socket_bind (nh,
+                                      (void *) un,
+                                      sizeof (struct sockaddr_un)))
       {
-        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR, "bind(for AF_UNIX)");
+        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                             "bind(AF_UNIX)");
         GNUNET_SCHEDULER_shutdown ();
         return;
       }
-      if (GNUNET_OK != GNUNET_NETWORK_socket_listen (nh, UNIX_BACKLOG))
+      if (GNUNET_OK !=
+          GNUNET_NETWORK_socket_listen (nh,
+                                        UNIX_BACKLOG))
       {
-        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR, "listen(for AF_UNIX)");
+        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                             "listen(AF_UNIX)");
         GNUNET_SCHEDULER_shutdown ();
         return;
       }
 
       fh = GNUNET_NETWORK_get_fd (nh);
-
-      if (0 != chmod (serve_unixpath, unixpath_mode))
+      GNUNET_NETWORK_socket_free_memory_only_ (nh);
+      if (0 != chmod (serve_unixpath,
+                      unixpath_mode))
       {
-        fprintf (stderr, "chmod failed: %s\n", strerror (errno));
+        GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                             "chmod");
         GNUNET_SCHEDULER_shutdown ();
         return;
       }
+      port = 0;
+    }
+    else if (0 == strcmp (serve_type, "tcp"))
+    {
+      char *bind_to;
 
-      mhd = MHD_start_daemon (MHD_USE_SUSPEND_RESUME,
-                              0,
-                              NULL, NULL,
-                              &url_handler, NULL,
-                              MHD_OPTION_LISTEN_SOCKET, fh,
-                              MHD_OPTION_NOTIFY_COMPLETED,
-                              &handle_mhd_completion_callback, NULL,
-                              MHD_OPTION_CONNECTION_TIMEOUT,
-                              (unsigned int) 10 /* 10s */,
-                              MHD_OPTION_END);
-      GNUNET_NETWORK_socket_free_memory_only_ (nh);
+      if (GNUNET_SYSERR ==
+          GNUNET_CONFIGURATION_get_value_number (config,
+                                                 "merchant",
+                                                 "PORT",
+                                                 &port))
+      {
+        GNUNET_log_config_missing (GNUNET_ERROR_TYPE_ERROR,
+                                   "merchant",
+                                   "PORT");
+        GNUNET_SCHEDULER_shutdown ();
+        return;
+      }
+      if (GNUNET_OK ==
+          GNUNET_CONFIGURATION_get_value_string (config,
+                                                 "merchant",
+                                                 "BIND_TO",
+                                                 &bind_to))
+      {
+        char port_str[6];
+        struct addrinfo hints;
+        struct addrinfo *res;
+        int ec;
+        struct GNUNET_NETWORK_Handle *nh;
+
+        GNUNET_snprintf (port_str,
+                         sizeof (port_str),
+                         "%u",
+                         (uint16_t) port);
+        memset (&hints, 0, sizeof (hints));
+        hints.ai_family = AF_UNSPEC;
+        hints.ai_socktype = SOCK_STREAM;
+        hints.ai_protocol = IPPROTO_TCP;
+        hints.ai_flags = AI_PASSIVE | AI_IDN;
+        if (0 !=
+            (ec = getaddrinfo (bind_to,
+                               port_str,
+                               &hints,
+                               &res)))
+        {
+          GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                      "Failed to resolve BIND_TO address `%s': %s\n",
+                      bind_to,
+                      gai_strerror (ec));
+          GNUNET_free (bind_to);
+          GNUNET_SCHEDULER_shutdown ();
+          return;
+        }
+        GNUNET_free (bind_to);
+
+        if (NULL == (nh = GNUNET_NETWORK_socket_create (res->ai_family,
+                                                        res->ai_socktype,
+                                                        res->ai_protocol)))
+        {
+          GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                               "socket");
+          freeaddrinfo (res);
+          GNUNET_SCHEDULER_shutdown ();
+          return;
+        }
+        if (GNUNET_OK !=
+            GNUNET_NETWORK_socket_bind (nh,
+                                        res->ai_addr,
+                                        res->ai_addrlen))
+        {
+          GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                               "bind");
+          freeaddrinfo (res);
+          GNUNET_SCHEDULER_shutdown ();
+          return;
+        }
+        freeaddrinfo (res);
+        if (GNUNET_OK !=
+            GNUNET_NETWORK_socket_listen (nh,
+                                          UNIX_BACKLOG))
+        {
+          GNUNET_log_strerror (GNUNET_ERROR_TYPE_ERROR,
+                               "listen");
+          GNUNET_SCHEDULER_shutdown ();
+          return;
+        }
+        fh = GNUNET_NETWORK_get_fd (nh);
+        GNUNET_NETWORK_socket_free_memory_only_ (nh);
+      }
+      else
+      {
+        fh = -1;
+      }
     }
     else
     {
@@ -753,10 +1096,18 @@ run (void *cls,
       GNUNET_assert (0);
     }
   }
-
+  mhd = MHD_start_daemon (MHD_USE_SUSPEND_RESUME | MHD_USE_DUAL_STACK,
+                          port,
+                          NULL, NULL,
+                          &url_handler, NULL,
+                          MHD_OPTION_LISTEN_SOCKET, fh,
+                          MHD_OPTION_NOTIFY_COMPLETED, &handle_mhd_completion_callback, NULL,
+                          MHD_OPTION_CONNECTION_TIMEOUT, (unsigned int) 10 /* 10s */,
+                          MHD_OPTION_END);
   if (NULL == mhd)
   {
-    GNUNET_break (0);
+    GNUNET_log (GNUNET_ERROR_TYPE_ERROR,
+                "Failed to launch HTTP service, exiting.\n");
     GNUNET_SCHEDULER_shutdown ();
     return;
   }
@@ -775,7 +1126,13 @@ run (void *cls,
 int
 main (int argc, char *const *argv)
 {
-  static const struct GNUNET_GETOPT_CommandLineOption options[] = {
+  struct GNUNET_GETOPT_CommandLineOption options[] = {
+
+    GNUNET_GETOPT_option_flag ('C',
+                               "connection-close",
+                               "force HTTP connections to be closed after each request",
+                               &TMH_merchant_connection_close),
+
     GNUNET_GETOPT_OPTION_END
   };
 
