@@ -11,7 +11,7 @@
   A PARTICULAR PURPOSE.  See the GNU Lesser General Public License for more details.
 
   You should have received a copy of the GNU Lesser General Public License along with
-  TALER; see the file COPYING.LGPL.  If not, If not, see
+  TALER; see the file COPYING.LGPL.  If not, see
   <http://www.gnu.org/licenses/>
 */
 /**
@@ -151,6 +151,7 @@ check_forbidden (struct TALER_MERCHANT_Pay *ph,
     GNUNET_JSON_spec_end()
   };
   unsigned int i;
+  int ret;
 
   if (GNUNET_OK !=
       GNUNET_JSON_parse (json,
@@ -165,11 +166,16 @@ check_forbidden (struct TALER_MERCHANT_Pay *ph,
     if (0 == memcmp (&ph->coins[i].coin_pub,
                      &coin_pub,
                      sizeof (struct TALER_CoinSpendPublicKeyP)))
-      return check_coin_history (&ph->coins[i],
-                                 history);
+    {
+      ret = check_coin_history (&ph->coins[i],
+                                history);
+      GNUNET_JSON_parse_free (spec);
+      return ret;
+    }
   }
   GNUNET_break_op (0); /* complaint is not about any of the coins
                           that we actually paid with... */
+  GNUNET_JSON_parse_free (spec);
   return GNUNET_SYSERR;
 }
 
@@ -235,6 +241,7 @@ handle_pay_finished (void *cls,
               (unsigned int) response_code);
   ph->cb (ph->cb_cls,
           response_code,
+	  TALER_JSON_get_error_code (json),
           json);
   TALER_MERCHANT_pay_cancel (ph);
 }
@@ -244,23 +251,21 @@ handle_pay_finished (void *cls,
  * Pay a merchant.  API for wallets that have the coin's private keys.
  *
  * @param ctx the execution loop context
- * @param exchange_uri URI of the exchange that the coins belong to
- * @param h_wire hash of the merchant’s account details
- * @param h_contract hash of the contact of the merchant with the customer
- * @param transaction_id transaction id for the transaction between merchant and customer
+ * @param merchant_uri base URI of the merchant's backend
+ * @param instance which merchant instance will receive this payment
+ * @param h_contract_terms hashcode of the proposal being paid
  * @param amount total value of the contract to be paid to the merchant
  * @param max_fee maximum fee covered by the merchant (according to the contract)
  * @param merchant_pub the public key of the merchant (used to identify the merchant for refund requests)
  * @param merchant_sig signature from the merchant over the original contract
  * @param timestamp timestamp when the contract was finalized, must match approximately the current time of the merchant
- * @param transaction_id transaction id for the transaction between merchant and customer
- * @param merchant_pub the public key of the merchant (used to identify the merchant for refund requests)
  * @param refund_deadline date until which the merchant can issue a refund to the customer via the merchant (can be zero if refunds are not allowed)
+ * @param pay_deadline maximum time limit to pay for this contract
+ * @param h_wire hash of the merchant’s account details
+ * @param exchange_uri URI of the exchange that the coins belong to
+ * @param order_id order id of the proposal being paid
  * @param num_coins number of coins used to pay
  * @param coins array of coins we use to pay
- * @param coin_sig the signature made with purpose #TALER_SIGNATURE_WALLET_COIN_DEPOSIT made by the customer with the coin’s private key.
- * @param max_fee maximum fee covered by the merchant (according to the contract)
- * @param amount total value of the contract to be paid to the merchant
  * @param pay_cb the callback to call when a reply for this request is available
  * @param pay_cb_cls closure for @a pay_cb
  * @return a handle for this request
@@ -268,16 +273,18 @@ handle_pay_finished (void *cls,
 struct TALER_MERCHANT_Pay *
 TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
 			   const char *merchant_uri,
-                           const struct GNUNET_HashCode *h_contract,
-                           uint64_t transaction_id,
+			   const char *instance,
+                           const struct GNUNET_HashCode *h_contract_terms,
 			   const struct TALER_Amount *amount,
 			   const struct TALER_Amount *max_fee,
                            const struct TALER_MerchantPublicKeyP *merchant_pub,
                            const struct TALER_MerchantSignatureP *merchant_sig,
                            struct GNUNET_TIME_Absolute timestamp,
                            struct GNUNET_TIME_Absolute refund_deadline,
+                           struct GNUNET_TIME_Absolute pay_deadline,
                            const struct GNUNET_HashCode *h_wire,
 			   const char *exchange_uri,
+                           const char *order_id,
                            unsigned int num_coins,
                            const struct TALER_MERCHANT_PayCoin *coins,
                            TALER_MERCHANT_PayCallback pay_cb,
@@ -287,13 +294,24 @@ TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
   struct TALER_DepositRequestPS dr;
   struct TALER_MERCHANT_PaidCoin pc[num_coins];
 
+  (void) GNUNET_TIME_round_abs (&timestamp);
+  (void) GNUNET_TIME_round_abs (&pay_deadline);
+  (void) GNUNET_TIME_round_abs (&refund_deadline);
+
+  if (GNUNET_YES !=
+      TALER_amount_cmp_currency (amount,
+                                 max_fee))
+  {
+    GNUNET_break (0);
+    return NULL;
+  }
+
   dr.purpose.purpose = htonl (TALER_SIGNATURE_WALLET_COIN_DEPOSIT);
   dr.purpose.size = htonl (sizeof (struct TALER_DepositRequestPS));
-  dr.h_contract = *h_contract;
+  dr.h_contract_terms = *h_contract_terms;
   dr.h_wire = *h_wire;
   dr.timestamp = GNUNET_TIME_absolute_hton (timestamp);
   dr.refund_deadline = GNUNET_TIME_absolute_hton (refund_deadline);
-  dr.transaction_id = GNUNET_htonll (transaction_id);
   dr.merchant = *merchant_pub;
   for (i=0;i<num_coins;i++)
   {
@@ -318,6 +336,13 @@ TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
     }
     TALER_amount_hton (&dr.deposit_fee,
 		       &fee);
+    {
+      TALER_LOG_DEBUG ("... amount_with_fee was %s\n",
+                       TALER_amount2s (&coin->amount_with_fee));
+      TALER_LOG_DEBUG ("... fee was %s\n",
+                       TALER_amount2s (&fee));
+    }
+
     GNUNET_CRYPTO_eddsa_sign (&coin->coin_priv.eddsa_priv,
 			      &dr.purpose,
 			      &p->coin_sig.eddsa_signature);
@@ -330,14 +355,8 @@ TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
   }
   return TALER_MERCHANT_pay_frontend (ctx,
 				      merchant_uri,
-				      h_contract,
-                                      amount,
-				      max_fee,
-				      transaction_id,
-				      merchant_sig,
-				      refund_deadline,
-				      timestamp,
-				      GNUNET_TIME_UNIT_ZERO_ABS,
+                                      merchant_pub,
+                                      order_id,
 				      exchange_uri,
 				      num_coins,
 				      pc,
@@ -349,21 +368,15 @@ TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
 /**
  * Pay a merchant.  API for frontends talking to backends. Here,
  * the frontend does not have the coin's private keys, but just
- * the public keys and signatures.  Note the sublte difference
+ * the public keys and signatures.  Note the subtle difference
  * in the type of @a coins compared to #TALER_MERCHANT_pay().
  *
  * @param ctx the execution loop context
+ * @param merchant_uri base URI of the merchant's backend
+ * @param merchant_pub public key of the merchant
  * @param exchange_uri URI of the exchange that the coins belong to
- * @param h_contract hash of the contact of the merchant with the customer
- * @param timestamp timestamp when the contract was finalized, must match approximately the current time of the merchant
- * @param transaction_id transaction id for the transaction between merchant and customer
- * @param refund_deadline date until which the merchant can issue a refund to the customer via the merchant (can be zero if refunds are not allowed)
- * @param wire_transfer_deadline date by which the merchant would like the exchange to execute the wire transfer (can be zero if there is no specific date desired by the frontend). If non-zero, must be larger than @a refund_deadline.
  * @param num_coins number of coins used to pay
  * @param coins array of coins we use to pay
- * @param coin_sig the signature made with purpose #TALER_SIGNATURE_WALLET_COIN_DEPOSIT made by the customer with the coin’s private key.
- * @param max_fee maximum fee covered by the merchant (according to the contract)
- * @param amount total value of the contract to be paid to the merchant
  * @param pay_cb the callback to call when a reply for this request is available
  * @param pay_cb_cls closure for @a pay_cb
  * @return a handle for this request
@@ -371,14 +384,8 @@ TALER_MERCHANT_pay_wallet (struct GNUNET_CURL_Context *ctx,
 struct TALER_MERCHANT_Pay *
 TALER_MERCHANT_pay_frontend (struct GNUNET_CURL_Context *ctx,
 			     const char *merchant_uri,
-                             const struct GNUNET_HashCode *h_contract,
-			     const struct TALER_Amount *amount,
-			     const struct TALER_Amount *max_fee,
-                             uint64_t transaction_id,
-                             const struct TALER_MerchantSignatureP *merchant_sig,
-                             struct GNUNET_TIME_Absolute refund_deadline,
-                             struct GNUNET_TIME_Absolute timestamp,
-                             struct GNUNET_TIME_Absolute wire_transfer_deadline,
+                             const struct TALER_MerchantPublicKeyP *merchant_pub,
+                             const char *order_id,
 			     const char *exchange_uri,
                              unsigned int num_coins,
                              const struct TALER_MERCHANT_PaidCoin *coins,
@@ -393,19 +400,6 @@ TALER_MERCHANT_pay_frontend (struct GNUNET_CURL_Context *ctx,
   struct TALER_Amount total_amount;
   unsigned int i;
 
-  if (GNUNET_YES !=
-      TALER_amount_cmp_currency (amount,
-                                 max_fee))
-  {
-    GNUNET_break (0);
-    return NULL;
-  }
-  if ( (0 != wire_transfer_deadline.abs_value_us) &&
-       (wire_transfer_deadline.abs_value_us < wire_transfer_deadline.abs_value_us) )
-  {
-    GNUNET_break (0);
-    return NULL;
-  }
   if (0 == num_coins)
   {
     GNUNET_break (0);
@@ -462,106 +456,30 @@ TALER_MERCHANT_pay_frontend (struct GNUNET_CURL_Context *ctx,
 			"ub_sig", GNUNET_JSON_from_rsa_signature (pc->denom_sig.rsa_signature),
 			"coin_sig", GNUNET_JSON_from_data_auto (&pc->coin_sig)
 			);
-    json_array_append (j_coins,
-		       j_coin);
+    GNUNET_assert (0 ==
+                   json_array_append_new (j_coins,
+                                          j_coin));
   }
 
-  { /* Sanity check that total_amount and total_fee
-       match amount/max_fee requirements */
-    struct TALER_Amount fee_left;
-
-    if (GNUNET_OK ==
-	TALER_amount_subtract (&fee_left,
-			       &total_fee,
-			       max_fee))
-    {
-      /* Wallet must cover part of the fee! */
-      struct TALER_Amount new_amount;
-
-      if (GNUNET_OK !=
-	  TALER_amount_add (&new_amount,
-			    &fee_left,
-			    amount))
-      {
-	/* integer overflow */
-	GNUNET_break (0);
-	json_decref (j_coins);
-	return NULL;
-      }
-      if (GNUNET_YES !=
-          TALER_amount_cmp_currency (&new_amount,
-                                     &total_amount))
-      {
-        GNUNET_break (0);
-	json_decref (j_coins);
-        return NULL;
-      }
-      if (1 ==
-	  TALER_amount_cmp (&new_amount,
-			    &total_amount))
-      {
-	/* new_amount > total_amount: all of the coins (total_amount)
-	   do not add up to at least the new_amount owed to the
-	   merchant, this request is bogus, abort */
-	GNUNET_break (0);
-	json_decref (j_coins);
-	return NULL;
-      }
-    }
-    else
-    {
-      /* Full fee covered by merchant, but our total
-	 must at least cover the total contract amount */
-      if (GNUNET_YES !=
-          TALER_amount_cmp_currency (amount,
-                                     &total_amount))
-      {
-        GNUNET_break (0);
-	json_decref (j_coins);
-        return NULL;
-      }
-      if (1 ==
-	  TALER_amount_cmp (amount,
-			    &total_amount))
-	{
-	  /* amount > total_amount: all of the coins (total_amount) do
-	   not add up to at least the amount owed to the merchant,
-	   this request is bogus, abort */
-	GNUNET_break (0);
-	json_decref (j_coins);
-	return NULL;
-      }
-    }
-  } /* end of sanity check */
-
-  pay_obj = json_pack ("{s:o," /* H_contract */
-                       " s:I, s:o," /* transaction id, timestamp */
-                       " s:o, s:s," /* refund_deadline, exchange */
-                       " s:o, s:o," /* coins, max_fee */
-                       " s:o, s:o}",/* amount, signature */
-                       "H_contract", GNUNET_JSON_from_data_auto (h_contract),
-                       "transaction_id", (json_int_t) transaction_id,
-                       "timestamp", GNUNET_JSON_from_time_abs (timestamp),
-                       "refund_deadline", GNUNET_JSON_from_time_abs (refund_deadline),
+  pay_obj = json_pack ("{"
+                       " s:s," /* exchange */
+                       " s:o," /* coins */
+                       " s:s," /* order_id */
+                       " s:o," /* merchant_pub */
+                       "}",
 		       "exchange", exchange_uri,
 		       "coins", j_coins,
-                       "max_fee", TALER_JSON_from_amount (max_fee),
-                       "amount", TALER_JSON_from_amount (amount),
-                       "merchant_sig", GNUNET_JSON_from_data_auto (merchant_sig));
-
-  if (0 != wire_transfer_deadline.abs_value_us)
-  {
-    /* Frontend did have an execution date in mind, add it */
-    json_object_set_new (pay_obj,
-			 "wire_transfer_deadline",
-			 GNUNET_JSON_from_time_abs (wire_transfer_deadline));
-  }
+                       "order_id", order_id,
+                       "merchant_pub", GNUNET_JSON_from_data_auto (merchant_pub));
 
   ph = GNUNET_new (struct TALER_MERCHANT_Pay);
   ph->ctx = ctx;
   ph->cb = pay_cb;
   ph->cb_cls = pay_cb_cls;
-  ph->url = GNUNET_strdup (merchant_uri);
+  GNUNET_asprintf (&ph->url,
+                   "%s%s",
+                   merchant_uri,
+                   "/pay");
   ph->num_coins = num_coins;
   ph->coins = GNUNET_new_array (num_coins,
                                 struct TALER_MERCHANT_PaidCoin);
@@ -609,6 +527,7 @@ TALER_MERCHANT_pay_cancel (struct TALER_MERCHANT_Pay *pay)
     GNUNET_CURL_job_cancel (pay->job);
     pay->job = NULL;
   }
+  GNUNET_free (pay->coins);
   GNUNET_free (pay->url);
   GNUNET_free (pay->json_enc);
   GNUNET_free (pay);
